#! /usr/bin/python
# -*- coding: utf-8 -*-

# HARK Middleware
# Copyright (C) 2013-2019 by Honda Research Institute Japan Co., Ltd.
# $Id:$

from harkmw.defs.terminal import TerminalDef
from harkmw.defs.parameter import Parameters
from harkmw.HarkError import HARKError

class NodeDef:
    """ネットワークファイル内のノード定義を管理する
    """

    def __init__(self, network, typedefs):
        """ノード定義を初期化する

        Parameters
        ----------
        network : NetworkDef
            このノードが属するネットワークの定義情報
        typedefs : Typedefs
            型情報を収めた Typedefs 型オブジェクト
        """
        self.network = network
        self.typedefs = typedefs
        self.name = ''
        self.type = ''
        self.inputs = {}
        self.outputs = {}
        self.parameters = Parameters()

        self.process = None
        self.topic = None

    def loadXML(self, element):
        """XML ノードからノード定義を取得する

        Parameters
        ----------
        element : xml.dom.minidom.Node
            ノード定義を取得する XML ノード
        """
        name = element.attributes['name'].value if element.getAttributeNode(
            'name') else None
        type = element.attributes['type'].value if element.getAttributeNode(
            'type') else None

        if name is None:
            raise HARKError("No attribute 'name' in element <%s>" %
                            element.nodeName)

        if type is None:
            raise HARKError("No attribute 'type' in element <%s>" %
                            element.nodeName)

        # attribute 'x' and 'y' are igonored
        parameters = Parameters()
        parameters.loadXML(element)

        document = self.network.document
        if document is not None:
            typedefs = document.typedefs.typedefs
            typedef = typedefs[type] if type in typedefs else None
        else:
            typedef = None

        inputs = {}
        outputs = {}

        if typedef is None:
            # ignore
            pass

        else:
            parameters.update(typedef.parameters)

            for (iname, input) in typedef.inputs.items():
                terminal = TerminalDef(self, input.name, True)
                inputs[iname] = terminal

            for (oname, output) in typedef.outputs.items():
                terminal = TerminalDef(self, output.name, False)
                outputs[oname] = terminal

        self.name = name
        self.type = type
        self.inputs = inputs
        self.outputs = outputs
        self.parameters = parameters

    def insert_params(self, parameters):
        """パラメータ値の上書きを行う

        Parameters
        ----------
        parameters : Parameters
            上書きする内容を収めた Parameters 型オブジェクト
        """
        self.parameters.update(parameters)

    def findInputTerminal(self, name):
        """指定した名前の入力端子を取得する

        Parameters
        ----------
        name : str
            端子の名前
        """
        from harkmw.defs.network import NetworkDef
        return self.findTerminal(NetworkDef.INPUT, name)

    def findOutputTerminal(self, name):
        """指定した名前の出力端子を取得する

        Parameters
        ----------
        name : str
            端子の名前
        """
        from harkmw.defs.network import NetworkDef
        return self.findTerminal(NetworkDef.OUTPUT, name)

    def findTerminal(self, type, name):
        """指定した名前の端子を取得する

        Parameters
        ----------
        name : str
            端子の名前
        """
        from harkmw.defs.network import NetworkDef
        if type == NetworkDef.INPUT:
            return self.inputs[name] if name in self.inputs else None
        elif type == NetworkDef.OUTPUT:
            return self.outputs[name] if name in self.outputs else None
        else:
            return None

    def addInputTerminal(self, name):
        """指定した名前の入力端子を追加する

        Parameters
        ----------
        name : str
            端子の名前
        """
        from harkmw.defs.network import NetworkDef
        return self.addTerminal(NetworkDef.INPUT, name)

    def addOutputTerminal(self, name):
        """指定した名前の出力端子を追加する

        Parameters
        ----------
        name : str
            端子の名前
        """
        from harkmw.defs.network import NetworkDef
        return self.addTerminal(NetworkDef.OUTPUT, name)

    def addTerminal(self, type, name):
        """指定した名前の入力端子を追加する

        Parameters
        ----------
        type : str
            端子の種類
        name : str
            端子の名前
        """
        from harkmw.defs.network import NetworkDef
        from harkmw.defs.terminal import TerminalDef
        if type == NetworkDef.INPUT:
            newTerminal = TerminalDef(self, name, True)
            self.inputs[name] = newTerminal
            return newTerminal

        elif type == NetworkDef.OUTPUT:
            newTerminal = TerminalDef(self, name, False)
            self.outputs[name] = newTerminal
            return newTerminal

        else:
            return None

    def get_process_name(self):
        """このノードが実行されるプロセスの名前を取得する
        """
        if self.parameters.has_parameter('HARKMW.PROCESS'):
            return self.parameters.get_parameter('HARKMW.PROCESS').value
        else:
            return self.network.get_process_name()

    def dump(self, prefix=''):
        """デバッグ用のダンプ出力
        """
        print(prefix + "[Node]")
        print(prefix + "  name=%s" % self.name)
        print(prefix + "  type=%s" % self.type)

        print(prefix + "  [Parameters]")
        self.parameters.dump(prefix + "  ")

# end of file
