#! /usr/bin/python
# -*- coding: utf-8 -*-

# HARK Middleware
# Copyright (C) 2013-2019 by Honda Research Institute Japan Co., Ltd.
# $Id:$

import copy
from harkmw.HarkError import HARKError

class Parameter:
    """ネットワークのパラメータを管理する
    """

    def __init__(self, name, type, value, description=None):
        """ネットワークのパラメータを初期化する

        Parameters
        ----------
        name : str
            パラメータ名
        type : str
            パラメータ種類
        value : str
            パラメータの値
        description : str
            パラメータの説明
        """
        self.name = name
        self.type = type
        self.value = value
        self.description = description
        self.used = False

    def __str__(self):
        """デバッグ用の文字列出力
        """
        return 'parameter: {}=({}){}'.format(self.name, self.type, self.value)


class Parameters:
    """ネットワークのパラメータの集合を管理する
    """

    def __init__(self):
        """パラメータの集合を初期化する
        """
        self.params = {}

    def clone(self):
        """パラメータの集合を複製する
        """
        new_object = Parameters()
        new_object.params = copy.deepcopy(self.params)
        return new_object

    def loadXML(self, node):
        """XML ノードからパラメータの集合を取得する

        Parameters
        ----------
        element : xml.dom.minidom.Node
            パラメータの集合を取得する XML ノード
        """
        for child in node.childNodes:
            if child.nodeType == child.ELEMENT_NODE:
                if child.nodeName == 'Parameter':
                    name = child.attributes['name'].value if child.getAttributeNode(
                        'name') else None
                    type = child.attributes['type'].value if child.getAttributeNode(
                        'type') else None
                    value = child.attributes['value'].value if child.getAttributeNode(
                        'value') else None
                    description = child.attributes['description'].value if child.getAttributeNode(
                        'description') else None

                    if not value and name in self.params:
                        value = self.params[name].value

                    self.params[name] = Parameter(
                        name, type, value, description)

                elif child.nodeName == 'Comments':
                    pass

                else:
                    pass

    def add(self, name, type, value, description=None):
        """パラメータを追加する

        Parameters
        ----------
        name : str
            パラメータ名
        type : str
            パラメータ種類
        value : str
            パラメータの値
        description : str
            パラメータの説明
        """
        self.params[name] = Parameter(name, type, value)

    def update(self, default_params):
        """パラメータの集合の既定値を設定する

        Parameters
        ----------
        default_params : Parameters
            既定値を収めた Parameters 型オブジェクト
        """
        for param in default_params.params.values():
            if param.name not in self.params:
                self.params[param.name] = param

        return self

    def normalize(self, network_parameters):
        """パラメータの集合の subnet_param を実際の値に置き換える

        Parameters
        ----------
        network_parameters : Parameters
            ネットワークのパラメータの集合を収めた Parameters 型オブジェクト
        """
        new_params = {}
        for param in self.params.values():
            if param.type == 'subnet_param':
                if network_parameters.has_parameter(param.value):
                    new_params[param.name] = network_parameters.get_parameter(
                        param.value)
                else:
                    raise HARKError(
                        "Parameter %s has type of %s, but network does not have parameter of the name" %
                        (param.name, param.type))

            else:
                new_params[param.name] = param
        self.params = new_params
        return self

    def has_parameter(self, name):
        """指定した名前のパラメータが定義されているか否かを返す

        Parameters
        ----------
        name : str
            パラメータ名
        """
        return name in self.params

    def get_parameter(self, name):
        """指定した名前のパラメータを返す

        Parameters
        ----------
        name : str
            パラメータ名
        """
        return self.params[name]

    def dump(self, prefix=''):
        """デバッグ用のダンプ出力
        """
        for name, param in self.params.items():
            print(
                prefix + '  [%s]: name="%s", type="%s", value="%s", description="%s"' %
                (name, param.name, param.type, param.value, param.description))

# end of file
