/* -*- mode: C++; coding: utf-8 -*- */

/*
 * Copyright (c) 2013 - 2020 Honda Research Institute Japan Co., Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include "libharkmw.h"

#include <pybind11/pybind11.h>
#include <pybind11/iostream.h>

#include <hark-base/Buffer.h>
#include <hark-base/Source.h>

#include "proxy.h"
#include "node.h"
#include "object.h"
#include "type.h"

#ifdef _WIN64
#include <windows.h>
#include <errno.h>

#define dlopen(P,G) (void*)LoadLibrary(P)
__inline const char* dlerror()
{
    static char szMsgBuf[256];
    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
                  NULL,
                  GetLastError(),
                  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                  szMsgBuf,
                  sizeof szMsgBuf,
                  NULL);
    return szMsgBuf;
}
#else
#include <dlfcn.h>
#endif

namespace HARKMW
{
void
load_dynamic(const std::string &path)
{
    if(dlopen(path.c_str(), RTLD_LAZY | RTLD_GLOBAL) == NULL) {
        std::stringstream ss;
        ss << "Failed to load shared library '" << path << "': " << dlerror();
        throw std::runtime_error(ss.str());
    }
}

void
print_modules()
{
    std::map<std::string, HARK::ModuleInfo>::iterator it = HARK::Object::module_info_dic.begin();
    while (it != HARK::Object::module_info_dic.end()) {
        std::stringstream ss;
        ss << it->first << " " << it->second.module_version;
        if (it->second.svn_version != "") {
            ss << " (Revision:" << it->second.svn_version << ")";
        }
        std::cout << ss.str() << std::endl;
        it++;
    }
}

hark_exception::hark_exception(const std::string &_message, HARK::BaseException *_base_exception)
    : std::runtime_error("")
{
    std::stringstream ss;
    ss << _message << ": ";
    _base_exception->print(ss);

    message = ss.str();
}

const char *
hark_exception::what() const noexcept
{
    return this->message.c_str();
}

}

namespace py = pybind11;

py::object print_as_hark_object(py::object x, py::object file)
{
    py::scoped_ostream_redirect stream(
        std::cout,
        py::isinstance<py::none>(file) ? py::module::import("sys").attr("stdout") : file
    );

    HARK::ObjectRef oref = HARKMW::fromPyObject(x);
    try {
        std::cout << oref;

    } catch(HARK::BaseException *ex) {
        std::cerr << "Exception in object printing: ";
        ex->print(std::cerr);
        std::cerr << std::endl;

    } catch(std::exception &ex) {
        std::cerr << "Exception in object printing: " << ex.what() << std::endl;

    } catch(...) {
        std::cerr << "Exception in object printing: " << "unknown" << std::endl;

    }

    py::object y;
    try {
        y = HARKMW::toPyObject(oref);

    } catch(HARK::BaseException *ex) {
        std::cerr << "Exception in object conversion to python: ";
        ex->print(std::cerr);
        std::cerr << std::endl;

    } catch(std::exception &ex) {
        std::cerr << "Exception in object conversion to python: " << ex.what() << std::endl;

    } catch(...) {
        std::cerr << "Exception in object conversion to python: " << "unknown" << std::endl;

    }

    return y;
}

PYBIND11_MODULE(harkmwnative, m)
{
    m.doc() = "harkmwnative: HARK C++ module wrapper";

    m.def("load_dynamic",  &HARKMW::load_dynamic);
    m.def("print_modules", &HARKMW::print_modules);

    m.def("print_as_hark_object", &print_as_hark_object, "", py::arg("x") = py::none(), py::arg("file") = py::none());

    py::class_<HARKMW::Node>(m, "Node")
    .def(py::init([](py::object node, const std::string &name, py::object params) {
        return new HARKMW::Node(node, name, params);
    }))
    .def("get_input_count", &HARKMW::Node::get_input_count)
    .def("get_output_count", &HARKMW::Node::get_output_count)
    .def("get_output", &HARKMW::Node::get_output)
    .def("connect", &HARKMW::Node::connect)
    .def("get_output_look_window", &HARKMW::Node::get_output_look_window)
    .def("get_input_look_window", &HARKMW::Node::get_input_look_window)
    .def("set_output_look_window", &HARKMW::Node::set_output_look_window)
    .def("set_input_look_window", &HARKMW::Node::set_input_look_window)
    .def("initialize", &HARKMW::Node::initialize)
    .def("process", &HARKMW::Node::process, py::arg("output_name"), py::arg("count"), py::arg("input_count"), py::arg("output_count"))
    .def("dump_input", &HARKMW::Node::dump_input)
    ;
    py::class_<HARKMW::Proxy>(m, "Proxy")
    .def(py::init([](py::object proxy) {
        return new HARKMW::Proxy(proxy);
    }))
    .def("connect_input", &HARKMW::Proxy::connect_input)
    .def("connect_output", &HARKMW::Proxy::connect_output)
    .def("initialize", &HARKMW::Proxy::initialize, py::arg("tag"))
    .def("push_data", &HARKMW::Proxy::push_data)
    .def("get_data", &HARKMW::Proxy::get_data)
    .def("set_data", &HARKMW::Proxy::set_data)
    .def("get_look_window", &HARKMW::Proxy::get_look_window)
    .def("set_look_window", &HARKMW::Proxy::set_look_window)
    .def("force_count", &HARKMW::Proxy::force_count)
    ;

    py::enum_<HARK::Source::CompareMode>(m, "CompareMode")
    .value("DEG", HARK::Source::CompareMode::DEG)
    .value("TFINDEX", HARK::Source::CompareMode::TFINDEX)
    .export_values();

    py::register_exception<HARKMW::node_not_implemented>(m, "NodeImplementationNotFoundError");
    py::register_exception<HARKMW::hark_exception>(m, "HarkError");
}
