/* -*- mode: C++; coding: utf-8 -*- */

/*
 * Copyright (c) 2013 - 2020 Honda Research Institute Japan Co., Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include "libharkmw.h"

#include <hark-base/BufferedNode.h>
#include <hark-base/Vector.h>
#include <hark-base/Matrix.h>
#include <hark-base/Map.h>

#include <hark-base/Source.h>
#include <hark-base/Tensor.h>
#include <hark-base/Time.h>
#include <hark-base/TimeStamp.h>

#include <hark-base/HarkParamsObjects.h>

#include "type.h"

#include <typeinfo>
#include <typeindex>
#include <pybind11/chrono.h>


namespace HARKMW
{
/**
Base converter class between HARK object and Python object.
use ClassObjectConverterBase
*/
template<class Derived>
class ClassObjectConverterBase
{
public:
    py::dict
    to_dict(HARK::ObjectRef &src)
    {
        typedef typename Derived::class_type C;
        C &obj = HARK::object_cast<C>(src);
        py::dict dict;
        dict["_type"] = this->get_class_name();
        static_cast<Derived *>(this)->map(obj, dict);
        return dict;
    }

    HARK::ObjectRef
    from_dict(py::dict &src)
    {
        HARK::ObjectRef odst(new typename Derived::class_type);
        typename Derived::class_type &obj = HARK::object_cast<typename Derived::class_type>(odst);

        assert(src["_type"].cast<std::string>() == this->get_class_name());
        static_cast<Derived *>(this)->unmap(src, obj);
        return odst;
    }

protected:
    const std::string &get_class_name() const
    {
        return static_cast<const Derived *>(this)->class_key;
    }

    py::handle assert_key(const py::dict &src, const char *key) const
    {
        if(src.contains(key)) {
            return src[key];
        } else {
            std::stringstream ss;
            ss << "trying to convert py::dict to class object, but dict not contain key '" << key << "', which is mandatory for class " << this->get_class_name();
            throw std::runtime_error(ss.str());
        }
    }

    template<typename T>
    T
    assert_key_cast(const py::dict &src, const char *key, const char *tname = nullptr) const
    {
        auto h = assert_key(src, key);
        try {
            T v = py::cast<T>(h);
            return v;
        } catch(py::cast_error &ex) {
            std::stringstream ss;
            ss << "trying to convert py::dict to " << this->get_class_name() << " object, but type of dict item for key '" << key << "' is not suitable, expected '" << (tname ? tname : typeid(T).name()) << "'";
            throw std::runtime_error(ss.str());
        }
    }
};

/**
Converter class between HARK object and Python object.

To add suppoted class type:
  - define static member class_key, which is used as '_type' value for python dict
  - implement specialized map/unmap method for C = classtype

*/
template<typename C>
class ClassObjectConverter : public ClassObjectConverterBase<ClassObjectConverter<C>>
{
public:
    using class_type = C;
    static std::string class_key;

public:
    void map  (const C &src, py::dict &dst) const;
    void unmap(const py::dict &src, C &dst) const;
};

/*
specialized map/unmap method for type HARK::Source
*/
template<>
std::string ClassObjectConverter<HARK::Source>::class_key = "Source";

template<>
void
ClassObjectConverter<HARK::Source>::map(const HARK::Source &src, py::dict &dst) const
{
    // std::cerr << "ClassObjectConverter<HARK::Source>::map" << std::endl;
    std::array<float, 3> arr_x           = {src.x[0], src.x[1], src.x[2]};
    std::array<int,   3> arr_tfindex     = {src.tfindex[0], src.tfindex[1], src.tfindex[2]};
    std::array<int,   3> arr_tfindex_max = {src.tfindex_max[0], src.tfindex_max[1], src.tfindex_max[2]};
    std::array<int,   3> arr_tfindex_min = {src.tfindex_min[0], src.tfindex_min[1], src.tfindex_min[2]};

    dst["x"] = arr_x;
    dst["power"] = src.power;
    dst["id"] = src.id;
    dst["remaining"] = src.remaining;
    dst["count"] = src.count;
    dst["last_found_count"] = src.last_found_count;
    dst["tfindex"] = arr_tfindex;
    dst["tfindex_max"] = arr_tfindex_max;
    dst["tfindex_min"] = arr_tfindex_min;
    dst["compare_mode"] = src.compare_mode;
}

template<>
void
ClassObjectConverter<HARK::Source>::unmap(const py::dict &src, HARK::Source &dst) const
{
    // std::cerr << "ClassObjectConverter<HARK::Source>::unmap" << std::endl;
    auto arr_x = assert_key_cast<std::array<float, 3>>(src, "x");
    auto arr_tfindex = assert_key_cast<std::array<int, 3>>(src, "tfindex");
    auto arr_tfindex_max = assert_key_cast<std::array<int, 3>>(src, "tfindex_max");
    auto arr_tfindex_min = assert_key_cast<std::array<int, 3>>(src, "tfindex_min");

    for(int i = 0; i < 3; i++) dst.x[i] = arr_x[i];
    dst.power = assert_key_cast<float>(src, "power");
    dst.id = assert_key_cast<int>(src, "id");
    dst.remaining = assert_key_cast<float>(src, "remaining");
    dst.count = assert_key_cast<int>(src, "count");
    dst.last_found_count = assert_key_cast<int>(src, "last_found_count");
    for(int i = 0; i < 3; i++) dst.tfindex[i] = arr_tfindex[i];
    for(int i = 0; i < 3; i++) dst.tfindex_max[i] = arr_tfindex_max[i];
    for(int i = 0; i < 3; i++) dst.tfindex_min[i] = arr_tfindex_min[i];

    dst.compare_mode = assert_key_cast<HARK::Source::CompareMode>(src, "compare_mode");
}

/*
specialized map/unmap method for type HARK::Tensor
*/
template<>
std::string ClassObjectConverter<HARK::Tensor>::class_key = "Tensor";

template<>
void
ClassObjectConverter<HARK::Tensor>::map(const HARK::Tensor &src, py::dict &dst) const
{
    dst["vsize"] = src.vsize;
    dst["data"]  = src.data;
}

template<>
void
ClassObjectConverter<HARK::Tensor>::unmap(const py::dict &src, HARK::Tensor &dst) const
{
    dst.vsize = assert_key_cast<std::vector<int>>(src, "vsize");
    dst.data = assert_key_cast<std::vector<float>>(src, "data");
}

/*
specialized map/unmap method for type HARK::Time
*/
template<>
std::string ClassObjectConverter<HARK::Time>::class_key = "Time";

template<>
void
ClassObjectConverter<HARK::Time>::map(const HARK::Time &src, py::dict &dst) const
{
    dst["time"] = src.time;
}

template<>
void
ClassObjectConverter<HARK::Time>::unmap(const py::dict &src, HARK::Time &dst) const
{
    dst.time = assert_key_cast<std::chrono::system_clock::time_point>(src, "time");
}

/*
specialized map/unmap method for type HARK::TimeStamp
*/
template<>
std::string ClassObjectConverter<HARK::TimeStamp>::class_key = "TimeStamp";

template<>
void
ClassObjectConverter<HARK::TimeStamp>::map(const HARK::TimeStamp &src, py::dict &dst) const
{
    dst["sec"] = src.time.tv_sec;
    dst["usec"] = src.time.tv_usec;
}

template<>
void
ClassObjectConverter<HARK::TimeStamp>::unmap(const py::dict &src, HARK::TimeStamp &dst) const
{
    dst.time.tv_sec = assert_key_cast<time_t>(src, "sec");
    dst.time.tv_usec = assert_key_cast<suseconds_t>(src, "usec");
}

/*
specialized map/unmap method for type ParamsLocalizeMUSIC
*/
template<>
std::string ClassObjectConverter<ParamsLocalizeMUSIC>::class_key = "ParamsLocalizeMUSIC";

template<>
void
ClassObjectConverter<ParamsLocalizeMUSIC>::map(const ParamsLocalizeMUSIC &src, py::dict &dst) const
{
    dst["num_source"] = src.num_source;
    dst["min_deg"] = src.min_deg;
    dst["max_deg"] = src.max_deg;
    dst["lower_bound_frequency"] = src.lower_bound_frequency;
    dst["upper_bound_frequency"] = src.upper_bound_frequency;
}

template<>
void
ClassObjectConverter<ParamsLocalizeMUSIC>::unmap(const py::dict &src, ParamsLocalizeMUSIC &dst) const
{
    dst.num_source = assert_key_cast<int>(src, "num_source");
    dst.min_deg = assert_key_cast<int>(src, "min_deg");
    dst.max_deg = assert_key_cast<int>(src, "max_deg");
    dst.lower_bound_frequency = assert_key_cast<int>(src, "lower_bound_frequency");
    dst.upper_bound_frequency = assert_key_cast<int>(src, "upper_bound_frequency");
}

/*
specialized map/unmap method for type ParamsSourceTracker
*/
template<>
std::string ClassObjectConverter<ParamsSourceTracker>::class_key = "ParamsSourceTracker";

template<>
void
ClassObjectConverter<ParamsSourceTracker>::map(const ParamsSourceTracker &src, py::dict &dst) const
{
    dst["thresh"] = src.thresh;
    dst["pause_length"] = src.pause_length;
    dst["min_src_interval"] = src.min_src_interval;
    dst["min_tfindex_interval"] = src.min_tfindex_interval;
    dst["compare_mode"] = src.compare_mode;
}

template<>
void
ClassObjectConverter<ParamsSourceTracker>::unmap(const py::dict &src, ParamsSourceTracker &dst) const
{
    dst.thresh = assert_key_cast<float>(src, "thresh");
    dst.pause_length = assert_key_cast<float>(src, "pause_length");
    dst.min_src_interval = assert_key_cast<float>(src, "min_src_interval");
    dst.min_tfindex_interval = assert_key_cast<float>(src, "min_tfindex_interval");
    dst.compare_mode = assert_key_cast<HARK::Source::CompareMode>(src, "compare_mode");
}

/*
specialized map/unmap method for type ParamsHRLE
*/
template<>
std::string ClassObjectConverter<ParamsHRLE>::class_key = "ParamsHRLE";

template<>
void
ClassObjectConverter<ParamsHRLE>::map(const ParamsHRLE &src, py::dict &dst) const
{
    dst["lx"] = src.lx;
    dst["time_constant"] = src.time_constant;
}

template<>
void
ClassObjectConverter<ParamsHRLE>::unmap(const py::dict &src, ParamsHRLE &dst) const
{
    dst.lx = assert_key_cast<float>(src, "lx");
    dst.time_constant = assert_key_cast<float>(src, "time_constant");
}

extern std::map<std::type_index, std::function<py::object(HARK::ObjectRef &)>> fromObjectRefTable;

//    extern std::function<HARK::ObjectRef(py::object)> findConverterFromPyObject(py::object obj);

py::object
Object2Null(HARK::ObjectRef &obj)
{
    return py::none();
}

template<typename H, typename C>
py::object
Object2Primitive(HARK::ObjectRef &obj)
{
    return py::cast(HARK::dereference_cast<C>(obj));
}

template<typename H, typename C>
py::object
Object2Object(HARK::ObjectRef &obj)
{
    return py::cast(HARK::object_cast<C>(obj));
}

template<typename C>
py::object
VectorObjectToList(HARK::ObjectRef &obj)
{
    return py::cast(HARK::object_cast<std::vector<C>>(obj));
}

template<typename C>
py::object
VectorObjectToNumpy(HARK::ObjectRef &obj)
{
    std::vector<C> o = HARK::object_cast<std::vector<C>>(obj);
    return py::array_t<C>({o.size()}, {sizeof(C)}, o.data());
}

template<typename C>
py::object
MatrixObjectToList(HARK::ObjectRef &obj)
{
    HARK::Matrix<C> &input_mat = HARK::object_cast<HARK::Matrix<C>>(obj);
    std::vector<std::vector<C>> mat(input_mat.nrows(), std::vector<C>(input_mat.ncols()));
    for(int r = 0; r < input_mat.nrows(); r++) {
        for(int c = 0; c < input_mat.ncols(); c++) {
            mat[r][c] = input_mat(r, c);
        }
    }
    return py::cast(mat);
}

template<typename C>
py::object
MatrixObjectToNumpy(HARK::ObjectRef &obj)
{
    HARK::Matrix<C> &o = HARK::object_cast<HARK::Matrix<C>>(obj);
    return py::array_t<C>({o.nrows(), o.ncols()}, {sizeof(C) * o.ncols(), sizeof(C)}, o[0]);
}

py::object
VectorObject2ObjectVector(HARK::ObjectRef &obj)
{
    HARK::Vector<HARK::ObjectRef> &vec = HARK::object_cast<HARK::Vector<HARK::ObjectRef>>(obj);
    if(vec.empty()) {
        return py::list();
    }

    py::list list = py::list(vec.size());
    for(size_t i = 0; i < vec.size(); i++) {
        list[i] = toPyObject(vec[i]);
    }

    return list;
}

py::object
MatrixObjectToObjectMatrix(HARK::ObjectRef &obj)
{
    HARK::Matrix<HARK::ObjectRef> mat = HARK::object_cast<HARK::Matrix<HARK::ObjectRef>>(obj);
    if(mat.ncols() == 0 || mat.nrows() == 0) {
        return py::list();
    }

    py::list mlist = py::list(mat.nrows());
    for(int r = 0; r < mat.nrows(); r++) {
        py::list vlist = py::list(mat.ncols());
        for(int c = 0; c < mat.ncols(); c++) {
            vlist[c] = toPyObject(mat(r, c));
        }
        mlist[r] = vlist;
    }

    return mlist;
}

py::object
MapObjectToDict(HARK::ObjectRef &obj)
{
    HARK::Map<int, HARK::ObjectRef> map = HARK::object_cast<HARK::Map<int, HARK::ObjectRef>>(obj);
    HARK::Map<int, HARK::ObjectRef>::iterator it = map.begin();

    auto pmap = std::map<int, py::object>();
    for(; it != map.end(); ++it) {
        pmap.emplace_hint(pmap.end(), it->first, toPyObject(it->second));
    }

    return py::cast(pmap);
}

template<typename C>
py::object
ClassObjectToDict(HARK::ObjectRef &obj)
{
    ClassObjectConverter<C> converter;
    return converter.to_dict(obj);
}

std::map<std::type_index, std::function<py::object(HARK::ObjectRef &)>> fromObjectRefTable = {
    {typeid(HARK::NilObject), Object2Null},
    {typeid(HARK::Bool), Object2Primitive<HARK::Bool, bool>},
    {typeid(HARK::Char), Object2Primitive<HARK::Char, char>},
    {typeid(HARK::Int), Object2Primitive<HARK::Int, int>},
    {typeid(HARK::Float), Object2Primitive<HARK::Float, float>},
    {typeid(HARK::Double), Object2Primitive<HARK::Double, double>},
    {typeid(HARK::String), Object2Object<HARK::String, std::string>},
    {typeid(HARK::Complex<float>), Object2Object<HARK::Complex<float>, std::complex<float>>},
    {typeid(HARK::Complex<double>), Object2Object<HARK::Complex<double>, std::complex<double>>},

    {typeid(HARK::Vector<bool>), VectorObjectToList<bool>},
    {typeid(HARK::Vector<char>), VectorObjectToList<char>},
    {typeid(HARK::Vector<int>), VectorObjectToNumpy<int>},
    {typeid(HARK::Vector<float>), VectorObjectToNumpy<float>},
    {typeid(HARK::Vector<double>), VectorObjectToNumpy<double>},
    {typeid(HARK::Vector<std::complex<float>>), VectorObjectToNumpy<std::complex<float>>},
    {typeid(HARK::Vector<std::complex<double>>), VectorObjectToNumpy<std::complex<double>>},
    {typeid(HARK::Vector<HARK::ObjectRef>), VectorObject2ObjectVector},

    {typeid(HARK::Matrix<bool>), MatrixObjectToList<bool>},
    {typeid(HARK::Matrix<char>), MatrixObjectToList<char>},
    {typeid(HARK::Matrix<int>), MatrixObjectToNumpy<int>},
    {typeid(HARK::Matrix<float>), MatrixObjectToNumpy<float>},
    {typeid(HARK::Matrix<double>), MatrixObjectToNumpy<double>},
    {typeid(HARK::Matrix<std::complex<float>>), MatrixObjectToNumpy<std::complex<float>>},
    {typeid(HARK::Matrix<std::complex<double>>), MatrixObjectToNumpy<std::complex<double>>},
    {typeid(HARK::Matrix<HARK::ObjectRef>), MatrixObjectToObjectMatrix},

    {typeid(HARK::Map<int, HARK::ObjectRef>), MapObjectToDict},

    {typeid(HARK::Source), ClassObjectToDict<HARK::Source>},
    {typeid(HARK::Tensor), ClassObjectToDict<HARK::Tensor>},
    {typeid(HARK::Time), ClassObjectToDict<HARK::Time>},
    {typeid(HARK::TimeStamp), ClassObjectToDict<HARK::TimeStamp>},
    {typeid(ParamsLocalizeMUSIC), ClassObjectToDict<ParamsLocalizeMUSIC>},
    {typeid(ParamsSourceTracker), ClassObjectToDict<ParamsSourceTracker>},
    {typeid(ParamsHRLE), ClassObjectToDict<ParamsHRLE>},
};

std::function<py::object(HARK::ObjectRef &)>
findConverterFromObjectRef(HARK::ObjectRef &obj)
{
    auto it = fromObjectRefTable.find(typeid(*obj));
    if(it != fromObjectRefTable.end()) {
        return it->second;
    } else {
        return nullptr;
    }
}

py::object
toPyObject(HARK::ObjectRef &obj)
{
    auto conv = findConverterFromObjectRef(obj);
    if(conv) {
        return conv(obj);
    } else {
        std::stringstream ss;
        ss << "Unsupported HARK type: " << typeid(*obj).name();
        throw std::runtime_error(ss.str());
    }
}

HARK::ObjectRef
PyObjectToNone(py::object obj)
{
//        std::cout << "PyObjectToNone" << std::endl;
    return HARK::nilObject;
}

template<typename PT, typename CT, typename HT>
HARK::ObjectRef
PyObjectToPrimitive(py::object obj)
{
    // std::cerr << "PyObjectToPrimitive: 1" << std::endl;
    auto v = obj.cast<PT>();
    HARK::ObjectRef o = HARK::ObjectRef(new HT((CT)v));
    return o;
}

template<typename PT, typename CT>
HARK::ObjectRef
PyObjectToPrimitiveList(py::list list)
{
//        std::cout << "PyObjectToPrimitiveList" << std::endl;
    std::vector<CT> cvec = list.cast<std::vector<CT>>();
    HARK::ObjectRef oref(new HARK::Vector<CT>(list.size()));
    HARK::Vector<CT> &hvec = *(static_cast<HARK::Vector<CT> *>(&(*oref)));;

    size_t i;
    for(i = 0; i < list.size(); i++) {
        hvec[i] = cvec[i];
    }

    return oref;
}

HARK::ObjectRef
PyObjectToObjectList(py::list list)
{
//        std::cout << "PyObjectToObjectList" << std::endl;
    HARK::ObjectRef oref(new HARK::Vector<HARK::ObjectRef>(list.size()));
    HARK::Vector<HARK::ObjectRef> &hvec = *(static_cast<HARK::Vector<HARK::ObjectRef> *>(&(*oref)));;

    size_t i;
    for(i = 0; i < list.size(); i++) {
        hvec[i] = fromPyObject(list[i]);
    }

    return oref;
}

template<typename PT, typename CT>
HARK::ObjectRef
PyObjectToPrimitiveMatrix(py::list list)
{
//        std::cout << "PyObjectToPrimitiveMatrix" << std::endl;
    size_t rows = list.size();
    py::list firstRow = py::cast<py::list>(list[0]);
    size_t cols = firstRow.size();

    std::vector<std::vector<CT>> cvec = list.cast<std::vector<std::vector<CT>>>();
    HARK::ObjectRef oref(new HARK::Matrix<CT>(rows, cols));
    HARK::Matrix<CT> &hvec = *(static_cast<HARK::Matrix<CT> *>(&(*oref)));;

    size_t i, j;
    for(i = 0; i < rows; i++) {
        py::list row = py::cast<py::list>(list[i]);
        if(row.size() != cols) {
            throw std::runtime_error("column size mismatch");
        }
        for(j = 0; j < cols; j++) {
            hvec[i][j] = cvec[i][j];
        }
    }

    return oref;
}

HARK::ObjectRef
PyObjectToObjectMatrix(py::list list)
{
//        std::cout << "PyObjectToObjectMatrix" << std::endl;
    py::list first = py::cast<py::list>(list[0]);
    size_t rows = list.size();
    size_t cols = first.size();

    HARK::ObjectRef oref(new HARK::Matrix<HARK::ObjectRef>(rows, cols));
    HARK::Matrix<HARK::ObjectRef> &hvec = *(static_cast<HARK::Matrix<HARK::ObjectRef> *>(&(*oref)));;

    size_t i, j;
    for(i = 0; i < rows; i++) {
        py::list row = py::cast<py::list>(list[i]);
        if(row.size() != cols) {
            throw std::runtime_error("column size mismatch");
        }
        for(j = 0; j < cols; j++) {
            hvec[i][j] = fromPyObject(row[j]);
        }
    }

    return oref;
}

HARK::ObjectRef
PyObjectToList(py::object obj)
{
//        std::cout << "PyObjectToList" << std::endl;
    py::list list = obj;
    if(list.size() == 0) {
        return HARK::ObjectRef(new HARK::Vector<HARK::ObjectRef>());

    } else {
        py::object first = list[0];

        if(py::isinstance<py::bool_>(first)) {
            return PyObjectToPrimitiveList<py::bool_, bool>(list);

        } else if(py::isinstance<py::int_>(first)) {
            return PyObjectToPrimitiveList<py::int_, int>(list);

        } else if(py::isinstance<py::float_>(first)) {
            return PyObjectToPrimitiveList<py::float_, float>(list);

        } else if(PyComplex_Check(first.ptr())) {
            return PyObjectToPrimitiveList<std::complex<float>, std::complex<float>>(list);

        } else if(py::isinstance<py::str>(first)) {
            return PyObjectToPrimitiveList<py::str, std::string>(list);

        } else {
            return PyObjectToObjectList(obj);

        }
    }
}

HARK::ObjectRef
PyObjectToMatrix(py::object obj)
{
//        std::cout << "PyObjectToMatrix" << std::endl;
    py::list list = obj;

    if(list.size() == 0) {
        return HARK::ObjectRef(new HARK::Matrix<HARK::ObjectRef>());

    } else {
        py::list firstRow = list[0];
        if(firstRow.size() == 0) {
            return HARK::ObjectRef(new HARK::Matrix<HARK::ObjectRef>(0, 0));
        }

        py::object first = firstRow[0];

        if(py::isinstance<py::bool_>(first)) {
            return PyObjectToPrimitiveMatrix<py::bool_, bool>(obj);

        } else if(py::isinstance<py::int_>(first)) {
            return PyObjectToPrimitiveMatrix<py::int_, int>(obj);

        } else if(py::isinstance<py::float_>(first)) {
            return PyObjectToPrimitiveMatrix<py::float_, float>(obj);

        } else if(PyComplex_Check(first.ptr())) {
            return PyObjectToPrimitiveMatrix<std::complex<float>, std::complex<float>>(obj);

        } else if(py::isinstance<py::str>(first)) {
            return PyObjectToPrimitiveMatrix<py::str, std::string>(obj);

        } else {
            return PyObjectToObjectMatrix(obj);
        }
    }
}

HARK::ObjectRef
PyDictToMap(py::object obj)
{
//        std::cout << "PyDictToMap" << std::endl;
    py::dict dict = obj;

    HARK::ObjectRef oref(new HARK::Map<int, HARK::ObjectRef>());
    HARK::Map<int, HARK::ObjectRef> &map = *(static_cast<HARK::Map<int, HARK::ObjectRef> *>(&(*oref)));

    for(auto it = dict.begin(); it != dict.end(); ++it) {
        int id = py::cast<py::int_>(it->first);
        HARK::ObjectRef value = fromPyObject(py::cast<py::object>(it->second));

        map[id] = value;
    }

    return oref;
}

HARK::ObjectRef
PyDictToClassObject(py::object obj)
{
    // std::cerr << "PyDictToClassObject" << std::endl;
    py::dict dict = obj;

    if(!dict.contains("_type")) {
        throw std::runtime_error("trying to convert py::dict to class object, but dict not contain key '_type'");
    }

    std::string type = dict["_type"].cast<std::string>();
    if(type == ClassObjectConverter<HARK::Source>::class_key) {
        ClassObjectConverter<HARK::Source> c;
        return c.from_dict(dict);

    } else if(type == ClassObjectConverter<HARK::Tensor>::class_key) {
        ClassObjectConverter<HARK::Tensor> c;
        return c.from_dict(dict);

    } else if(type == ClassObjectConverter<HARK::Time>::class_key) {
        ClassObjectConverter<HARK::Time> c;
        return c.from_dict(dict);

    } else if(type == ClassObjectConverter<HARK::TimeStamp>::class_key) {
        ClassObjectConverter<HARK::TimeStamp> c;
        return c.from_dict(dict);

    } else if(type == ClassObjectConverter<ParamsLocalizeMUSIC>::class_key) {
        ClassObjectConverter<ParamsLocalizeMUSIC> c;
        return c.from_dict(dict);

    } else if(type == ClassObjectConverter<ParamsSourceTracker>::class_key) {
        ClassObjectConverter<ParamsSourceTracker> c;
        return c.from_dict(dict);

    } else if(type == ClassObjectConverter<ParamsHRLE>::class_key) {
        ClassObjectConverter<ParamsHRLE> c;
        return c.from_dict(dict);

    } else {
        std::stringstream ss;
        ss << "trying to convert py::dict to class object, but dict contains unexpected value '" << type << "' for key '_type'";
        throw std::runtime_error(ss.str());
    }

    return HARK::nilObject;
}

template<typename CT>
HARK::ObjectRef
PyArrayToList(py::object obj)
{
//        std::cout << "PyArrayToList" << std::endl;

    py::array_t<CT> arr = obj;
    const auto &arr_info = arr.request();
    const auto &arr_shape = arr_info.shape;

    size_t size = arr_shape[0];

    HARK::ObjectRef oref(new HARK::Vector<CT>(size));
    HARK::Vector<CT> &hvec = *(static_cast<HARK::Vector<CT> *>(&(*oref)));;

    size_t i;
    for(i = 0; i < size; i++) {
        hvec[i] = *arr.data(i);
    }

    return oref;
}

template<typename CT>
HARK::ObjectRef
PyArrayToMatrix(py::object obj)
{
//        std::cout << "PyArrayToMatrix" << std::endl;

    py::array_t<CT> arr = obj;
    const auto &arr_info = arr.request();
    const auto &arr_shape = arr_info.shape;
    // size_t ndim = arr_shape.size();

    size_t rows = arr_shape[0];
    size_t cols = arr_shape[1];

    HARK::ObjectRef oref(new HARK::Matrix<CT>(rows, cols));
    HARK::Matrix<CT> &hvec = *(static_cast<HARK::Matrix<CT> *>(&(*oref)));;

    size_t i, j;
    for(i = 0; i < rows; i++) {
        for(j = 0; j < cols; j++) {
            hvec[i][j] = *arr.data(i, j);
        }
    }

    return oref;
}

std::function<HARK::ObjectRef(py::object)>
findConverterFromPyObject(py::object obj)
{
//        std::cout << "findConverterFromPyObject" << std::endl;
    if(py::isinstance<py::none>(obj)) {
        return PyObjectToNone;

    } else if(py::isinstance<py::bool_>(obj)) {
        return PyObjectToPrimitive<py::bool_, bool, HARK::Bool>;

    } else if(py::isinstance<py::int_>(obj)) {
        return PyObjectToPrimitive<py::int_, int, HARK::Int>;

    } else if(py::isinstance<py::float_>(obj)) {
        return PyObjectToPrimitive<py::float_, float, HARK::Float>;

    } else if(PyComplex_Check(obj.ptr())) {
        return PyObjectToPrimitive<std::complex<float>, std::complex<float>, HARK::Complex<float>>;

    } else if(py::isinstance<py::str>(obj)) {
        return PyObjectToPrimitive<py::str, std::string, HARK::String>;

    } else if(py::isinstance<py::list>(obj)) {
        auto list = py::cast<py::list>(obj);
        if(list.size() > 0 && py::isinstance<py::list>(list[0])) {
            return PyObjectToMatrix;
        } else {
            return PyObjectToList;
        }

    } else if(py::isinstance<py::dict>(obj)) {
        py::dict dict = obj;
        if(dict.empty()) {
            return PyDictToMap;

        } else {
            auto it = dict.begin();
            if(py::isinstance<py::int_>(it->first)) {
                return PyDictToMap;

            } else {
                return PyDictToClassObject;

            }
        }

    } else if(py::isinstance<py::array>(obj)) {
        py::array arr = obj;

        const auto &arr_info = arr.request();
        const auto &arr_shape = arr_info.shape;
        size_t ndim = arr_shape.size();

        if(ndim > 2 || ndim < 1) {
            std::stringstream ss;
            ss << "Unexpected numpy ndim: " << ndim;
            throw std::runtime_error(ss.str());

        } else if(py::isinstance<py::array_t<std::int16_t>>(arr)
                  || py::isinstance<py::array_t<std::int32_t>>(arr)
                  || py::isinstance<py::array_t<std::int64_t>>(arr)
                  || py::isinstance<py::array_t<std::uint16_t>>(arr)
                  || py::isinstance<py::array_t<std::uint32_t>>(arr)
                  || py::isinstance<py::array_t<std::uint64_t>>(arr)) {
            if(ndim == 1) {
                return PyArrayToList<int>;
            } else {
                return PyArrayToMatrix<int>;
            }

        } else if(py::isinstance<py::array_t<float>>(arr)) {
            if(ndim == 1) {
                return PyArrayToList<float>;
            } else {
                return PyArrayToMatrix<float>;
            }

        } else if(py::isinstance<py::array_t<double>>(arr)) {
            if(ndim == 1) {
                return PyArrayToList<double>;
            } else {
                return PyArrayToMatrix<double>;
            }

        } else if(py::isinstance<py::array_t<std::complex<float>>>(arr)) {
            if(ndim == 1) {
                return PyArrayToList<std::complex<float>>;
            } else {
                return PyArrayToMatrix<std::complex<float>>;
            }

        } else if(py::isinstance<py::array_t<std::complex<double>>>(arr)) {
            if(ndim == 1) {
                return PyArrayToList<std::complex<double>>;
            } else {
                return PyArrayToMatrix<std::complex<double>>;
            }

        } else {
            std::stringstream ss;
            ss << "Unexpected numpy dtype: " << arr.dtype();
            throw std::runtime_error(ss.str());
        }

    } else {
        return nullptr;
    }
}

HARK::ObjectRef
fromPyObject(py::object obj)
{
    auto conv = findConverterFromPyObject(obj);
    if(conv) {
        return conv(obj);
    } else {
        std::stringstream ss;
        ss << "Unsupported python type: " << (std::string)py::str(obj.get_type());
        throw std::runtime_error(ss.str());
    }
}
}

/* end of file */
