#! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (C) 2013 - 2020 Honda Research Institute Japan Co., Ltd.
# All rights reserved.
# HARK was developed by researchers in Okuno Laboratory
# at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
# 
# Licensed under the HARK license, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# https://www.hark.jp/notice/HARK_License_Agreement.pdf

from harkmw.defs.parameter import Parameters

import harkmw.logic.node
import harkmwnative
from collections import defaultdict


class NativeNode(harkmw.logic.node.Node):
    """C++ で実装された HARK ノードをラップするクラス
    """

    def __init__(self):
        """クラスの初期化を行う
        """
        harkmw.logic.node.Node.__init__(self)
        self.network = None
        self.count = -1
        self.input_count = -1
        self.output_count = -1
        self.input_look_windows = defaultdict(lambda: (0, 0))
        self.output_look_windows = defaultdict(lambda: (0, 0))

    def build(self, network, node_def, params, force_local):
        """HARK ノードを構築する

        Parameters
        ----------
        network : Network
            構築するノードが属するネットワーク
        node_def : NodeDef
            構築するノードの定義情報
        params : Parameters
            構築するノードのパラメータ
        force_local : bool
            設定値によらず、local プロセスでの実行を想定する場合に True を指定する
        """
        self.network = network
        self.type = node_def.type
        self.name = node_def.name
        self.topic = node_def.topic
        self.process_name = 'local' if force_local else node_def.get_process_name()

        # Load native module
        self.native_node = harkmwnative.Node(
            self, node_def.type, node_def.parameters)

    def build_type(self, network, type, force_local, name=None):
        """ノード種類を指定して、HARK ノードを構築する

        Parameters
        ----------
        network : Network
            構築するノードが属するネットワーク
        type : str
            構築するノードの種類
        force_local : bool
            設定値によらず、local プロセスでの実行を想定する場合に True を指定する
        name : str
            ノードの名前
        """
        # Load native module
        self.native_node = harkmwnative.Node(self, type, Parameters())

        self.network = network
        self.type = type
        self.name = name
        self.topic = None
        self.process_name = 'local' if force_local else network.process_name

    def initialize(self, process):
        """ノードの初期化を行う。

        HARK ノードの initialize メソッドを呼び出し、必要な初期化を実行する

        Parameters
        ----------
        process : str
            プロセス名
        """
        self.native_node.initialize()
        harkmw.logic.node.Node.initialize(self, process)

    def release(self):
        """ノードを開放する
        """
        self.native_node = None
        self.inputs = {}
        self.outputs = {}

    def process(self, output, count, input_count=-1, output_count=-1):
        """HARK ノードに処理を実行させる

        Parameters
        ----------
        output : str
            出力端子名
        count : int
            フレーム番号を表す count 値を指定する
        input_count : int
            subnet の入出力端子を担うノード場合に、ノードの入力側の count 値を指定する
        output_count : int
            subnet の入出力端子を担うノード場合に、ノードの出力側の count 値を指定する
        """
        self.native_node.process(
            output, count, input_count=input_count, output_count=output_count)
        self.count = count

    def set_output_look_window(self, output_id, look_window):
        """HARK ノードの出力端子に lookAhead, lookBack を設定する

        Parameters
        ----------
        output_id : str
            出力端子名
        look_window : tuple
            lookAhead, lookBack の 2 値からなる tuple
        """
        self.output_look_windows[output_id] = look_window
        return self.native_node.set_output_look_window(
            output_id, look_window[0], look_window[1])

    def get_output_look_window(self, output_id):
        """HARK ノードの出力端子から lookAhead, lookBack を取得する

        Parameters
        ----------
        output_id : str
            出力端子名
        """
        look_window = self.native_node.get_output_look_window(output_id)
        return max(look_window, self.output_look_windows[output_id])

    def set_input_look_window(self, input_id, look_window):
        """HARK ノードの入力端子に lookAhead, lookBack を設定する

        Parameters
        ----------
        input_id : str
            入力端子名
        look_window : tuple
            lookAhead, lookBack の 2 値からなる tuple
        """
        self.input_look_windows[input_id] = look_window
        return self.native_node.set_input_look_window(
            input_id, look_window[0], look_window[1])

    def get_input_look_window(self, input_id):
        """HARK ノードの入力端子から lookAhead, lookBack を取得する

        Parameters
        ----------
        input_id : str
            入力端子名
        """
        look_window = self.native_node.get_input_look_window(input_id)
        return max(look_window, self.input_look_windows[input_id])

    def get_input_count(self):
        """HARK ノードの入力端子の個数を取得する
        """
        return self.native_node.get_input_count()

    def get_output_count(self):
        """HARK ノードの出力端子の個数を取得する
        """
        return self.native_node.get_output_count()

    def get_output(self, name, count):
        """HARK ノードの指定した出力を、Python 型に変換して取得する

        Parameters
        ----------
        name : str
            出力端子名
        count : int
            フレーム番号
        """
        return self.native_node.get_output(name, count)

    def dump_input(self, count):
        """デバッグ用のダンプ出力
        """
        return self.native_node.dump_input(count)

    def connect_node(self, terminal_name, node_from, terminal_from_name):
        """指定した HARK ノードと接続する

        Parameters
        ----------
        terminal_name : str
            接続する端子の名前
        node_from : NativeNode
            接続元となるノード
        terminal_from_name : str
            接続元の端子名
        """
        self.native_node.connect(
            terminal_name, node_from.native_node, terminal_from_name)

    def dump(self):
        """デバッグ用のダンプ出力
        """
        print('[NativeNode]')
        print('  type: %s' % self.type)
        print('  name: %s' % self.name)
        print('  network: type=%s, name=%s' %
              (self.network.type, self.network.name))

        print('  inputs: ')
        for (name, terminal) in self.inputs.items():
            print(
                '    Input: name=%s, node=%s/%s, terminal=%s' %
                (name,
                 terminal.input_node.type,
                 terminal.input_node.name,
                 terminal.input_terminal_name))

        print('  outputs: ')
        for (name, proxy_list) in self.outputs.items():
            for proxy in proxy_list:
                print(
                    '    Output: name=%s, node=%s/%s, terminal=%s' %
                    (name,
                     proxy.output_node.type,
                     proxy.output_node.name,
                     proxy.output_terminal_name))

    def __str__(self):
        """デバッグ用のダンプ出力
        """
        return "<%s: type=%s, name=%s>" % (
            self.__class__.__name__, self.type, self.name)

# end of file
