#!/usr/bin/env python3


from setuptools import setup, find_packages

try:
    from pybind11.setup_helpers import Pybind11Extension
except ImportError:
    from setuptools import Extension as Pybind11Extension

try:
    from pybind11.setup_helpers import build_ext
except ImportError:
    from setuptools.command.build_ext import build_ext

import os
import sys
import struct


__version__ = '3.6.0'

##
#@brief Get include_dirs for the python module.
#@details Support path acquisition for Python Extension module
# in distutils obsoleted environment.
#
#@return    A dictionary to set in the Extension() .
# e.g.) If you run try_get_python_path() on Ubuntu 23.04.
# {
#     'include': '/usr/include/python3.11',
#     'lib': '/usr/local/lib/python3.11/dist-packages'
# }
def try_get_python_path():
    if sys.version_info >= (3, 10):
        import sysconfig

        ## [for workaround]
        ## Workaround for versions that return /usr/local/include when
        ## the actual path uses /usr/include in Debian environment.
        scheme = sysconfig.get_default_scheme()
        if scheme == 'posix_local':
            scheme = 'posix_prefix'

        return {
            'include': sysconfig.get_path('platinclude', scheme),
            'lib': sysconfig.get_path('platlib')
        }
    else:
        from distutils import sysconfig

        return {
            'include': sysconfig.get_python_inc(plat_specific=True),
            'lib': sysconfig.get_python_lib(plat_specific=True)
        }

#print(try_get_python_path())


##
#@brief Get include_dirs for the python module.
#@details Enabled to get the path even if the version of 'pybind11' installed
# by the 'python3-pybind11' package is older than 2.6.0.
#
#@return    A dictionary to set in the Extension() .
# e.g.) If you run try_get_pybind11_path() on Ubuntu 23.04.
# {
#     'include': '/usr/lib/python3/dist-packages/pybind11/include'
# }
def try_get_pybind11_path():
    import pybind11

    retval = { 'include': "" }

    try:
        ## [for test]
        ## If get_include() didn't exist in the pybind11 module (Ver.2.5.x or earlier),
        ## an AttributeError exception would be raised.
        #raise AttributeError()

        retval["include"] = pybind11.get_include()
        return retval

    except AttributeError:
        import subprocess

        def pip_show():
            pip_prog = os.environ.get("PIP", None) or "pip"
            pip_proc = subprocess.Popen("{} show pybind11".format(pip_prog).split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = pip_proc.communicate()
            return out.rstrip().decode("utf-8")

        for location in pip_show().split("\n"):
            if location.startswith("Location: ") and len(location.split()) > 1:
                retval["include"] = "{}/pybind11/include".format(location.split()[1])
                break

        return retval

    return retval

#print(try_get_pybind11_path())


##
#@brief Get CFLAGS and LIBS for the packages.
#@details Do something like the python3-pkgconfig module
# if there is no python3-pkgconfig module.
#
#@param packages    A string of package names separated by whitespace.
#@param inherited_val    Append to an existing dict() or Extension().
#@return    A dictionary to set in the Extension() .
# e.g.) If you run try_pkgconfig("hark-base libharkio3") on Ubuntu 23.04.
# {
#     'extra_comple_args': ['-I/usr/include/hark', '-I/usr/include/hark/hark-base', '-I/usr/include/libxml2'],
#     'extra_link_args': ['-lhark-base', '-L/usr/lib/hark/toolbox/hark-base', '-lhark-base.tlb', '-lharkio3', '-lxml2', '-lzip'],
#     'include_dirs': ['/usr/include/hark', '/usr/include/hark/hark-base', '/usr/include/libxml2'],
#     'library_dirs': ['/usr/lib/hark/toolbox/hark-base'],
#     'libraries': ['hark-base', 'hark-base.tlb', 'harkio3', 'xml2', 'zip']
#     'runtime_library_dirs': [],
# }
def try_pkgconfig(packages, inherited_val=None):
    retval = dict()
    retval.setdefault('extra_compile_args', [])
    retval.setdefault('extra_link_args', [])
    retval.setdefault('include_dirs', [])
    retval.setdefault('library_dirs', [])
    retval.setdefault('libraries', [])
    retval.setdefault('runtime_library_dirs', [])

    try:
        ## [for test]
        ## If pkgconfig module isn't installed,
        ## an ImportError exception would be raised.
        #raise ImportError()

        import pkgconfig

        for package in packages.split():
            if not pkgconfig.exists(package):
                if(package == "pybind11"):
                    ## [for alternatives]
                    ## If the 'pybind11' package is not found,
                    ## try to get the headers path from the 'python3-pybind11' package,
                    ## or any 'pip install pybind11' package.
                    pybind11_inc_path = try_get_pybind11_path()["include"]
                    if pybind11_inc_path:
                        retval["extra_compile_args"].append("-I{}".format(pybind11_inc_path))
                    else:
                        raise EnvironmentError("pkg-config could not find the \"{}\".".format(package))
                else:
                    raise EnvironmentError("pkg-config could not find the \"{}\".".format(package))
            else:
                retval["extra_compile_args"].extend(pkgconfig.cflags(package).split())
                retval["extra_link_args"].extend(pkgconfig.libs(package).split())

    except ImportError:
        import subprocess

        def exists(_package):
            pkgconfig_prog = os.environ.get("PKG_CONFIG", None) or "pkg-config"
            return subprocess.call("{} --exists {}".format(pkgconfig_prog, _package).split()) == 0

        def cflags(_package):
            # Microsoft VisualStudio supports
            msvc_opts = "--msvc-syntax" if os.name == "nt" else ""
            pkgconfig_prog = os.environ.get("PKG_CONFIG", None) or "pkg-config"
            pkgconfig_proc = subprocess.Popen("{} --cflags {} {}".format(pkgconfig_prog, msvc_opts, _package).split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = pkgconfig_proc.communicate()
            return out.rstrip().decode("utf-8")

        def libs(_package):
            # Microsoft VisualStudio supports
            msvc_opts = "--msvc-syntax" if os.name == "nt" else ""
            pkgconfig_prog = os.environ.get("PKG_CONFIG", None) or "pkg-config"
            pkgconfig_proc = subprocess.Popen("{} --libs {} {}".format(pkgconfig_prog, msvc_opts, _package).split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = pkgconfig_proc.communicate()
            return out.rstrip().decode("utf-8")

        for package in packages.split():
            if not exists(package):
                raise EnvironmentError("pkg-config could not find the \"{}\".".format(package))
            retval["extra_compile_args"].extend(cflags(package).split())
            retval["extra_link_args"].extend(libs(package).split())

    retval["include_dirs"].extend([x[2:] for x in retval["extra_compile_args"] if x.startswith("-I") or x.startswith("/I")])
    retval["library_dirs"].extend([x[2:] for x in retval["extra_link_args"] if x.startswith("-L") or x.startswith("/L")])
    retval["libraries"].extend([x[2:] for x in retval["extra_link_args"] if x.startswith("-l") or x.startswith("/l")])
    # RPATH or RUNPATH should not be set.
    # https://lintian.debian.org/tags/custom-library-search-path
    # https://wiki.debian.org/RpathIssue
    #if os.name != "nt":
    #    retval["runtime_library_dirs"].extend(retval["library_dirs"])

    if isinstance(inherited_val, Pybind11Extension):
        retval['extra_compile_args'].extend(inherited_val.extra_compile_args)
        retval['extra_link_args'].extend(inherited_val.extra_link_args)
        retval['include_dirs'].extend(inherited_val.include_dirs)
        retval['library_dirs'].extend(inherited_val.library_dirs)
        retval['libraries'].extend(inherited_val.libraries)
        retval['runtime_library_dirs'].extend(inherited_val.runtime_library_dirs)

        inherited_val.extra_compile_args = retval["extra_compile_args"]
        inherited_val.extra_link_args = retval["extra_link_args"]
        inherited_val.include_dirs = retval["include_dirs"]
        inherited_val.library_dirs = retval["library_dirs"]
        inherited_val.libraries = retval["libraries"]
        inherited_val.runtime_library_dirs = retval["runtime_library_dirs"]
        return inherited_val

    elif isinstance(inherited_val, dict):
        retval['extra_compile_args'].extend(inherited_val.get('extra_compile_args', []))
        retval['extra_link_args'].extend(inherited_val.get('extra_link_args', []))
        retval['include_dirs'].extend(inherited_val.get('include_dirs', []))
        retval['library_dirs'].extend(inherited_val.get('library_dirs', []))
        retval['libraries'].extend(inherited_val.get('libraries', []))
        retval['runtime_library_dirs'].extend(inherited_val.get('runtime_library_dirs', []))

    return retval

## test codes
## Case.1 Set a simple library
#ext = try_pkgconfig("hark-base")
## Case.2 Set a dependent library was dependent few libraries
#ext = try_pkgconfig("glibmm-2.4")
## Case.3 Set the few libraries
#ext = try_pkgconfig("hark-base libharkio3 glibmm-2.4")
## Case.4 Set inherited dict()
#ext = try_pkgconfig("hark-base", {"dummy": True})
## Case.5 Set inherited Extension()
#ext_base = Pybind11Extension("foo", ["foo.cpp"])
#ext = try_pkgconfig("hark-base libharkio3 glibmm-2.4", ext_base)
#print(ext_base) # for check Extension() object address
#print(ext) # for check Extension() object address (same as previous object 'ext_base')

## Check Extension()'s attributes or dict()'s items.
#if isinstance(ext, Pybind11Extension):
#    print("ext.extra_compile_args:\n    {}".format(ext.extra_compile_args))
#    print("ext.extra_link_args:\n    {}".format(ext.extra_link_args))
#    print("ext.include_dirs:\n    {}".format(ext.include_dirs))
#    print("ext.library_dirs:\n    {}".format(ext.library_dirs))
#    print("ext.libraries:\n    {}".format(ext.libraries))
#elif isinstance(ext, dict):
#    print("ext['extra_compile_args']:\n    {}".format(ext['extra_compile_args']))
#    print("ext['extra_link_args']:\n    {}".format(ext['extra_link_args']))
#    print("ext['include_dirs']:\n    {}".format(ext['include_dirs']))
#    print("ext['library_dirs']:\n    {}".format(ext['library_dirs']))
#    print("ext['libraries']:\n    {}".format(ext['libraries']))

ext_modules = [
    try_pkgconfig("pybind11 hark-base", Pybind11Extension(
        "harkmwnative",
        # Sort input source files to ensure bit-for-bit reproducible builds
        # (https://github.com/pybind/python_example/pull/53)
        sources = sorted([
            'harkmw/src/node.cc',
            'harkmw/src/libmain.cc',
            'harkmw/src/proxy.cc',
            'harkmw/src/type.cc',
            'harkmw/src/collector.cc',
            ]),
        include_dirs = [
            try_get_python_path()["include"],
            'harkmw/src'],
        #library_dirs=[hb_libdir],
        #runtime_library_dirs=[hb_libdir] if os.name != 'nt' else None,
        #libraries=[hb_libname],
        language = "c++",
        extra_compile_args = ["-g", "-O0"],
    )),
]

# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile
    import os
    with tempfile.NamedTemporaryFile('w', suffix='.cpp', delete=False) as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        fname = f.name
    try:
        compiler.compile([fname], extra_postargs=[flagname])
    except setuptools.distutils.errors.CompileError:
        return False
    finally:
        try:
            os.remove(fname)
        except OSError:
            pass
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14/17] compiler flag.
    The newer version is prefered over c++11 (when it is available).
    """
    flags = ['-std=c++17', '-std=c++14', '-std=c++11']

    for flag in flags:
        if has_flag(compiler, flag):
            return flag

    raise RuntimeError('Unsupported compiler -- at least C++11 support '
                       'is needed!')

class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    c_opts = {
        'msvc': ['/EHsc'],
        'unix': [],
    }
    l_opts = {
        'msvc': [],
        'unix': [],
    }

    if sys.platform == 'darwin':
        darwin_opts = ['-stdlib=libc++', '-mmacosx-version-min=10.7']
        c_opts['unix'] += darwin_opts
        l_opts['unix'] += darwin_opts

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        link_opts = self.l_opts.get(ct, [])
        if ct == 'unix':
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, '-fvisibility=hidden'):
                opts.append('-fvisibility=hidden')

        for ext in self.extensions:
            ext.define_macros = [('VERSION_INFO', '"{}"'.format(self.distribution.get_version()))]
            ext.extra_compile_args = opts
            ext.extra_link_args = link_opts
        build_ext.build_extensions(self)

setup(
    name='harkmw',
    version=__version__,
    author='',
    author_email='',
    url='',
    description='',
    long_description='',
    data_files=[
        ('/var/lib/harkmw',
         [])],
    ext_modules=ext_modules,
    packages=find_packages(),
    install_requires=[
        'pybind11>=2.4',
        'paho-mqtt>=1.3',
    ],
    cmdclass={'build_ext': BuildExt},
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'harkmw=harkmw.main:app_main',
            'harkmwd=harkmw.main:daemon_main',
            'batchflow=harkmw.main:app_main']},
)

